#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import signal
import time

import numpy as np
import uuid
from ..bridge import server as ebs
from ..bridge import protocol as bp

from easydict import EasyDict as edict


class ProtocolStack(ebs.StreamFactory):
    def build_protocol_stack(self, s):
        s.add_input_protocol(bp.LengthSplitIn())
        s.add_input_protocol(bp.ImageDecoder())
        s.add_output_protocol(bp.ImageEncoder(quality=60))
        s.add_output_protocol(bp.LengthSplitOut())


class ImageServer(ebs.EaterBridgeServer):
    def __init__(self, **kargs):
        kargs["protocol_stack"] = ProtocolStack
        super().__init__(**kargs)
        self.data_queue = []
        self.model = None

    def update(self):
        if self.model is None:
            import evaluator
            self.model = evaluator.Evaluator()
        self.data_queue += self.getDataBlocksAsArray()
        if len(self.data_queue) > 0:
            batch_data, socket_mapper, self.data_queue = ebs.slice_as_batch_size(self.data_queue, 128)

            batch_data = np.uint8(batch_data)
            for i in range(len(batch_data)):
                img = batch_data[i]
                r_image = self.model.eval(img)
                batch_data[i] = r_image

            stored_datablocks = ebs.pack_array_datablock(socket_mapper, batch_data)
            self.setDataBlocksFromArray(stored_datablocks)



#!/usr/bin/env python3
from __future__ import print_function
import importlib
import os


def to_bool(s):
    return s in [1, "True", "TRUE", "true", "1", "yes", "Yes", "Y", "y", "t", "on"]


DEBUG = False
if "DEBUG" in os.environ:
    DEBUG = to_bool(os.environ["DEBUG"])
    if DEBUG:
        try:
            import __builtin__
        except ImportError:
            import builtins as __builtin__
        import inspect

        def lpad(s, c):
            return s[0:c].ljust(c)

        def rpad(s, c):
            if len(s) > c:
                return s[len(s) - c:]
            else:
                return s.rjust(c)

        def print(*args, **kwargs):
            s = inspect.stack()
            __builtin__.print("\033[44m%s@%s(%s):\033[0m " % (rpad(s[1][1], 20), lpad(str(s[1][3]), 10), rpad(str(s[1][2]), 4)), end="")
            return __builtin__.print(*args, **kwargs)


def _pre_():
    print("\033[A                                                                \033[A", flush=True)


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


_dopen = open



#!/usr/bin/env python3
import logging
import math
import os
import platform
import queue
import sys
import threading
import uuid
import multiprocessing

import twisted.internet.protocol
import twisted.internet.reactor

if platform.platform() == "Darwin":
    os.environ["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"

logger = logging.getLogger(__name__)
if __name__ == "__main__":
    import protocol as bridge_protocol
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter("%(asctime)s [%(filename)s:%(lineno)d] %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = True
else:
    from . import protocol as bridge_protocol
    logger.addHandler(logging.NullHandler())
    logger.setLevel(logging.DEBUG)
    logger.propagate = True


def estimate_retry_delay_time(r, max_delay=10):
    s = math.exp(r / 1.5)
    return s if s < 10 else 10


class StackedClientSocketProtocol(twisted.internet.protocol.Protocol):
    def __init__(self, rq, wq):
        self.rq = rq
        self.wq = wq
        self.input_middlewares = []
        self.output_middlewares = []
        self.is_available = False
        self.is_invalid_socket = False
        self.uuid = str(uuid.uuid4())

    def add_input_protocol(self, p):
        p.queue_name = self.uuid
        self.input_middlewares.append(p)

    def add_output_protocol(self, p):
        p.queue_name = self.uuid
        self.output_middlewares.append(p)

    def connectionMade(self):
        import aimage
        if aimage.is_native:
            cores = multiprocessing.cpu_count()
            aimage.rebuild_worker(cores)
            aimage.create_queue(self.uuid)
        self.is_available = True

    def connectionLost(self, reason):
        import aimage
        if aimage.is_native:
            aimage.delete_queue(self.uuid)
        self.is_available = False

    def dataReceived(self, data):
        #logger.info("TCP:READ:", data)
        if self.is_available:
            self.input_middlewares[0].write(data)

    def update(self):
        try:
            if self.is_invalid_socket: return
            if self.is_available is False: return
            try:
                # From opponent
                for i in range(len(self.input_middlewares) - 1):
                    b = self.input_middlewares[i].read()
                    if len(b):
                        self.input_middlewares[i + 1].write(b)
                for m in self.input_middlewares:
                    m.update()
                # To opponent
                for i in range(len(self.output_middlewares) - 1):
                    b = self.output_middlewares[i].read()
                    if len(b):
                        self.output_middlewares[i + 1].write(b)
                for m in self.output_middlewares:
                    m.update()
                buf = self.output_middlewares[-1].read(-1)
                if self.is_available and len(buf) > 0:
                    self.transport.write(buf)
                    #logger.info("TCP:WRITE:", buf)
            except Exception as e:
                logger.warning(e)
                self.is_invalid_socket = True
                try:
                    self.transport.close()
                    self.wq.clear()
                    self.rq.clear()
                except Exception as e:
                    logger.warning(e)
            while self.rq.empty() is False:
                o = self.rq.get_nowait()
                self.output_middlewares[0].write(o)

            b = self.input_middlewares[-1].read(-1)
            if len(b) > 0:
                self.wq.put_nowait(b)
        except Exception as e:
            logger.warning(e)


class StreamClientFactory(twisted.internet.protocol.ClientFactory):
    def __init__(self, rq, wq, **kwargs):
        self.rq = rq
        self.wq = wq
        self.retry = 0
        self.retying = False
        self.connected = False
        self.protocol_instance = None
        self.addr = None
        self.kwargs = kwargs
        self.update()

    def startedConnecting(self, connector):
        logger.info(f'Started to connect. {connector.host}:{connector.port} Timeout:{connector.timeout}')

    # Override
    def on_disconnected(self):
        pass

    # Override
    def on_connected(self):
        self.protocol_instance.add_input_protocol(bridge_protocol.DirectStream())
        self.protocol_instance.add_output_protocol(bridge_protocol.DirectStream())

    def buildProtocol(self, addr):
        self.addr = addr
        self.retry = 0
        self.retying = False
        self.connected = True
        logger.info(f'Connected {addr.type}://{addr.host}:{addr.port}')
        s = StackedClientSocketProtocol(self.rq, self.wq)
        self.protocol_instance = s
        self.on_connected()
        return s

    def update(self):
        if self.protocol_instance:
            self.protocol_instance.update()
            if self.protocol_instance.is_invalid_socket:
                self.protocol_instance = None
        twisted.internet.reactor.callLater(0.001, self.update)

    def deferred_connect(self, args):
        if self.retying is False and self.connected is False:
            self.retying = True
            self.retry += 1
            args[0].connect()

    def clientConnectionLost(self, connector, reason):
        delay = estimate_retry_delay_time(self.retry)
        logger.warning(f'Lost connection. RetryDelay:[{self.retry}]:{round(delay,2)}s\n  Reason: {reason}')
        self.on_disconnected()
        self.connected = False
        self.retying = False
        twisted.internet.reactor.callLater(delay, self.deferred_connect, (connector, ))

    def clientConnectionFailed(self, connector, reason):
        delay = estimate_retry_delay_time(self.retry)
        logger.warning(f'Connection failed. RetryDelay:[{self.retry}]:{round(delay,2)}s\n  Reason: {reason}')
        self.on_disconnected()
        self.connected = False
        self.retying = False
        twisted.internet.reactor.callLater(delay, self.deferred_connect, (connector, ))


class EaterBridgeClient:
    def __init__(self, **kargs):
        self.kargs = kargs
        self.host = kargs["host"]
        self.port = kargs["port"]
        self.rq = queue.Queue()
        self.wq = queue.Queue()
        self.protocol_stack = kargs["protocol_stack"](self.rq, self.wq, **self.kargs)

    def start(self):
        self.deferred = twisted.internet.reactor.connectTCP(self.host, self.port, self.protocol_stack)
        # self.deferred = twisted.internet.reactor.connectSSL(self.host, self.port, StreamClientFactory(self.rq,self.wq),twisted.internet.ssl.ClientContextFactory())
        self.thread = threading.Thread(target=twisted.internet.reactor.run, args=(False, ))
        self.thread.setDaemon(True)
        self.thread.start()

    def destroy(self):
        twisted.internet.reactor.stop()
        # self.thread.stop()

    def write(self, blocks):
        if self.rq.empty():
            self.rq.put(blocks)
            return True
        return False

    def read(self, size=-1):
        if self.wq.empty() is False:
            return self.wq.get()
        return None


def echo(HOST, PORT):
    class ProtocolStack(bridge.client.StreamClientFactory):
        def on_connected(self):
            s = self.protocol_instance
            s.add_input_protocol(bridge.protocol.DirectStream())
            s.add_output_protocol(bridge.protocol.DirectStream())

        def on_disconnected(self):
            pass

    c = bridge.client.EaterBridgeClient(host=HOST, port=PORT, protocol_stack=ProtocolStack)
    c.start()

    def terminate(a, b):
        c.destroy()
        exit(9)

    signal.signal(signal.SIGINT, terminate)
    signal.signal(signal.SIGTERM, terminate)
    request_count = 0
    index = 0
    while True:
        if request_count < 1:
            s = str(index)
            c.write(s)
            request_count += len(s)
            index += 1
        data = c.read()
        if isinstance(data, bytes) or isinstance(data, bytearray):
            logger.warning(data)
            request_count -= len(data)
        else:
            time.sleep(0.01)


def data2data(HOST, PORT):
    class ProtocolStack(bridge.client.StreamClientFactory):
        def on_connected(self):
            s = self.protocol_instance
            s.add_input_protocol(bridge.protocol.LengthSplitIn())
            s.add_output_protocol(bridge.protocol.LengthSplitOut())

        def on_disconnected(self):
            pass

    c = bridge.client.EaterBridgeClient(host=HOST, port=PORT, protocol_stack=ProtocolStack)
    c.start()

    def terminate(a, b):
        c.destroy()
        exit(9)

    signal.signal(signal.SIGINT, terminate)
    signal.signal(signal.SIGTERM, terminate)
    request_count = 0
    fps_count = 0
    st = time.time()
    while True:
        if request_count < 2:
            if time.time() - st > 1.0:
                logger.info("FPS: " + fps_count)
                st = time.time()
                fps_count = 0
            fps_count += 1

            datas = ["test", "test2"]
            logger.info(f"Main:send({request_count}): " + datas)
            c.write(datas)
            request_count += 2
        blocks = c.read()
        if blocks is not None:
            if isinstance(blocks, list):
                for data in blocks:
                    request_count -= 1
                    # TODO list / bytes
                    logger.info(f"Main:response({request_count}): " + data.decode('utf-8'))
                    # for data in extend:
                    #     print(data.decode('utf-8'))
                    # blocks = client.read()
                    # for img in blocks:
                    #     aimage.show(img)
        else:
            time.sleep(0.01)



#!/usr/bin/env python3
_dopen = open
import glob
import json
import os
import pathlib
import random
import sys
import time

import cv2
import numpy as np
import tqdm
from aimage.head import *
from aimage.img import *
from aimage.ui import *
from easydict import EasyDict as edict


# Stub
class AggressiveImageGenerator:
    def __init__(self, **kwargs):
        """
        kwargs:
         - verbose: True / False # display detail
         - shuffle: True / False # shuffle data
         - entry: <input data path>
         - label_path: <output label path> for resume
         - loss: "tree" / "list" 
         - target_size: (w,h,c) # shape
         - data_align: True / False # adjust data length for each classes
         - rescale: 1/255.0
         - data_aug_params: dict{}
         - progress_bar: True / False
         - batch_size: 128
        """

        entry = kwargs["entry"]
        if entry == "/":
            print("\033[0;31m", "Error: Can't use root:", entry, "\033[0m")
            raise Exception("Can't specify root path")
        if entry[-1] == '/':
            print("\033[0;31m", entry, "\033[0m")
            entry = entry[0:-1]
        if entry == ".":
            print("\033[0;31m", "Error: Current directory:", entry, "\033[0m")
            raise Exception("Invalid path")
        if os.path.exists(entry) is False:
            print("\033[0;31m", "Error: Does not exist path:", entry, "\033[0m")
            raise Exception("Does not exist path")

        self.loss = "list"
        self.verbose = False
        self.progress_bar = True
        self.pbar = None
        self.shuffle = False
        self.batch_size = 128
        self.label_path = "output.label"
        self.target_size = (256, 256, 3)
        self.rescale = 1 / 255.0
        self.data_aug_params = {"resize_width": 256, "resize_height": 256}

        self.set(**kwargs)

        if self.verbose:
            print(kwargs)

        self.q = 0
        self.iindex = 0
        if self.verbose:
            print("=============================================")
        self.oindex = 0
        self.STREAM_BATCH = 16

        self.label_name = os.path.basename(self.entry)
        self.build_classes()

        if self.verbose:
            print(self.classes)

        data_aug_params = kwargs["data_aug_params"]

        def setd(k, v):
            if k not in data_aug_params:
                data_aug_params[k] = v

        setd("resize_interpolation", "fastest")
        for k in data_aug_params:
            data_aug_params[k] = str(data_aug_params[k])
        self.data_aug_params = data_aug_params

        self.stub_buffer = []
        self.output_buffer = []
        self.datas = None
        self.is_tree = self.loss == "tree"
        self.sync_reset()

    def shape(self):
        return self.target_size

    def length(self):
        return self.total

    def set_description(self, s):
        if self.pbar:
            self.pbar.set_description(s)
        pass

    def __iter__(self):
        self.ite = 0
        if self.pbar:
            self.pbar.refresh()
            self.pbar.close()
            self.pbar = None
        if self.progress_bar:
            self.pbar = tqdm.tqdm(total=self.length(), leave=True, file=sys.stdout)
        return self

    def __next__(self):
        if self.total > self.ite:
            ret = self.get_batch(self.batch_size)
            dlen = len(ret.images)
            if self.pbar:
                self.pbar.update(dlen)
            self.ite += dlen
            ret.ite = self.ite
            return ret
        else:
            if self.pbar:
                self.pbar.refresh()
                self.pbar.close()
                self.pbar = None
            self.sync_reset()
            raise StopIteration()

    def sync_reset(self):
        entry = self.entry
        if self.datas is None:
            self.datas = []
            self.prebatch = None
            class_index_table = {}
            table = {}
            for clazz in self.classes:
                c = self.classes[clazz]
                class_index_table[c["index"]] = {"name": clazz, "results": [], "total": 0}
            for filename in pathlib.Path(entry).glob('**/*.jpg'):
                filename = str(filename)
                self.datas.append(filename)
                signals = self.make_signal(self.entry, filename, self.classes)
                if self.is_tree:
                    signals *= self.signal_mask
                idx_key = np.argmax(signals)
                if idx_key not in table:
                    table[idx_key] = []
                table[idx_key].append(filename)
                # if idx_key not in class_index_table:
                #    class_index_table[idx_key] = {"name":self.get_name(idx_key),"results":[],"total":0}
                class_index_table[idx_key]["total"] += 1
            vmax = 0
            for k in table:
                v = table[k]
                vlen = len(v)
                if vlen > vmax:
                    vmax = vlen
                if self.verbose:
                    print(self.get_name(k), ":", vlen)
            if self.verbose:
                print("Max => ", vmax)
            for k in table:
                v = table[k]
                i = 0
                vlen = len(v)
                ns = []
                for x in range(vmax):
                    ns.append(v[x % vlen])
                table[k] = ns
            new_datas = []
            for k in table:
                v = table[k]
                vlen = len(v)
                new_datas += v
            if self.data_align:
                if self.verbose:
                    print("Total", len(self.datas), "=>", vmax, "x", len(table), "=>", len(new_datas))
                self.datas = new_datas
                for i in class_index_table:
                    class_index_table[i]["total"] = vmax
            self.total = len(self.datas)
            self.class_index_table = class_index_table
        if self.shuffle:
            random.shuffle(self.datas)
        self.iindex = 0
        self.oindex = 0

    def build_classes(self):
        j = {}
        try:
            j = json.loads(_dopen(self.label_path).read())
        except:
            pass
        classes = self.make_class(entry=self.entry, loss=self.loss, class_dict=j)
        self.label_json = json.dumps(classes)

        if os.path.exists(os.path.dirname(self.label_path)) is False:
            os.makedirs(os.path.dirname(self.label_path), exist_ok=True)

        with _dopen(self.label_path, "w") as fp:
            fp.write(self.label_json)

        self.classes = classes
        if self.loss == "tree":
            signal_mask = np.zeros((len(classes), ))
            tcnt = 0
            for clazz in classes:
                i = classes[clazz]["index"]
                if clazz[0] == "@":
                    signal_mask[i] = 1
                    tcnt += 1
            if tcnt <= 2:
                raise "Two or more target directories are required. Use @ to define it."
            self.signal_mask = signal_mask
        else:
            self.signal_mask = np.ones((len(classes), ))
        return classes

    def get_classes(self):
        return self.clasess

    def get_batch(self, batch_size=None):
        if batch_size is None:
            batch_size = self.batch_size
        block = self.get_data_block(batch_size)
        b = edict()
        b.images = []
        b.signals = []
        b.points = None
        b.file_paths = []
        if block:
            for d in block:
                b.images.append(d["image"])
                b.signals.append(d["signals"])
                b.file_paths.append(d["file_path"])
                #d["points"])
            b.images = np.array(b.images, dtype=np.float32)
            b.signals = np.array(b.signals, dtype=np.float32)
            if self.rescale != 1.0:
                b.images *= self.rescale

        return b

    def get_data_block(self, batch_size):
        if self.total == 0:
            print("Zero length")
            raise "Zero length"
        while True:
            if self.q < 1024:
                if self.iindex < self.total:
                    ds = self.datas[self.iindex:self.iindex + self.STREAM_BATCH]
                    dlen = len(ds)
                    self.q += dlen
                    stream = list()
                    # input   = dict()
                    # input["data_aug_params"] = self.data_aug_params
                    for image_path in ds:
                        signals = self.make_signal(self.entry, image_path, self.classes)
                        d = dict()
                        d["image_path"] = image_path
                        d["file_path"] = image_path
                        d["image"] = cv2.resize(load_image(image_path), (self.target_size[0], self.target_size[1]), interpolation=cv2.INTER_AREA)

                        # data_aug_params
                        # {'entry': 'data/fruit/train', 'label_path': 'weights/fruit.mobilenet.categorical_crossentropy.label', 'loss': 'categorical_crossentropy',
                        #     'target_size': (224, 224, 3), 'data_align': True, 'rescale': 0.00392156862745098, 'shuffle': True, 'data_aug_params': {'resize_width': 224, 'resize_height': 224}}

                        d["signals"] = np.array(signals, dtype=np.float32)
                        d["points_table"] = None
                        stream.append(d)
                    # input["stream"] = stream
                    # //    {
                    # //      params:{
                    # //          data_aug_params:{}
                    # //      },
                    # //      stream: [{image_path:"",signals:[],points_table:[]},{},{},{},]
                    # //    }
                    self.stub_buffer += stream
                    # native_module.async_image_loader_with_data_aug_input(str(hex(id(self))),input)
                    self.iindex += dlen
                else:
                    # random.shuffle(self.datas)
                    # self.iindex = 0
                    break
            else:
                break
        if self.iindex == self.oindex:
            return None
        if len(self.output_buffer) >= batch_size or self.iindex == self.oindex + len(self.output_buffer):
            buf = self.output_buffer[0:batch_size]
            self.output_buffer = self.output_buffer[batch_size:]
            self.oindex += len(buf)
            return buf
        while True:
            #ret = native_module.async_image_loader_with_data_aug_output(str(hex(id(self))))
            ret = self.stub_buffer
            if ret is not None and len(ret) > 0:
                self.q -= len(ret)
                self.output_buffer += ret
                if len(self.output_buffer) >= batch_size or self.iindex == self.oindex + len(self.output_buffer):
                    buf = self.output_buffer[0:batch_size]
                    self.output_buffer = self.output_buffer[batch_size:]
                    self.oindex += len(buf)
                    return buf
            else:
                time.sleep(0.1)
                if self.iindex == self.oindex:
                    return None

    def find_index(self, class_dict):
        index = 0
        d = sorted(class_dict.items(), key=lambda x: x[1]["index"])
        for obj in d:
            if index == int(obj[1]["index"]):
                index += 1
                continue
        return index

    def make_signal(self, entry, fpath, classes):
        sp = str(fpath)[len(entry):].split("/")
        signal = np.zeros((len(classes), ), dtype=np.float32)
        for s in sp:
            if s in classes:
                signal[classes[s]["index"]] = 1
        return signal

    def register_class(self, class_name, class_dict, mark):
        if class_name not in class_dict:
            index = self.find_index(class_dict)
            class_dict[class_name] = {"index": index}

    def make_class(self, entry, *, loss="tree", class_dict={}):
        if self.verbose:
            print("Entrypoint => ", entry)
        class_table_by_index = {}
        class_table_by_name = {}
        elen = len(entry)
        if self.verbose:
            print(loss)
        if loss == "list":
            class_names = [os.path.basename(f) for f in glob.glob(os.path.join(entry, "*")) if os.path.isdir(f)]
            # print(class_names)
            for class_name in class_names:
                self.register_class(class_name, class_dict, "S")
        elif loss == "tree":
            for filename in pathlib.Path(entry).glob('**/'):
                filename = str(filename)
                # {}Attributs/{@}Target/{!}Ignore/{?}Unsupervised
                if "/!" in filename:
                    print("Ignore", filename)
                    continue
                if filename == entry:
                    continue
                class_name = os.path.basename(filename)
                if class_name not in class_dict:
                    if class_name[0] == '@':
                        self.register_class(class_name, class_dict, mark="S")
                    elif class_name[0] == '?':
                        self.register_class(class_name, class_dict, mark=" ")
                    else:
                        self.register_class(class_name, class_dict, mark="-")
        else:
            raise "Inbalid loss"
        return class_dict

    def get_name(self, index):
        for k in self.classes:
            if index == self.classes[k]["index"]:
                return k
        return None

    def find_by_name(self, p, label_name):
        if label_name in self.classes:
            return p[self.classes[label_name]["index"]]
        return None

    def find_top(self, p):
        index = np.argmax(p)
        for k in self.classes:
            c = self.classes[k]
            if c["index"] == index:
                return index, p[index], k
        return None, None, None

    def set(self, **kwargs):
        for k in kwargs:
            setattr(self, k, kwargs[k])

    @staticmethod
    def make_full_aug_params():
        class Dict:
            def __init__(self):
                self.d = dict()

            def set(self, k, v):
                self.d[k] = str(v)

        d = Dict()
        d.set("random_gaussian", 0.1)
        d.set("random_sharp", 0.1)
        d.set("random_median", 0.1)
        d.set("random_bilateral", 0.1)
        d.set("random_mosaic", 0.1)
        d.set("mosaic_shift_range", 1)
        d.set("random_equalization", 0.1)
        d.set("random_color_reduction", 0.1)
        d.set("random_cos", 0.1)
        d.set("random_hsv", 0.5)
        d.set("hue_range", 15)
        d.set("saturation_range", 0.05)
        d.set("lightness_range", 0.1)
        d.set("random_pow", 0.1)
        d.set("pow_min", 0.9)
        d.set("pow_max", 1.3)
        d.set("random_horizontal_flip", 0.15)
        d.set("random_vertical_flip", 0.25)
        d.set("random_shuffle_splitted_images", 0.1)
        d.set("random_resize_and_arrange_images", 0.1)
        d.set("random_crop", 0.25)
        d.set("crop_range", 0.2)
        d.set("random_noise", 0.1)
        d.set("random_erase", 0.2)
        d.set("background_type", 5)
        d.set("affine_transform", 0.5)
        d.set("aggressive_transform", 0.5)
        d.set("random_x_shift", 0.1)
        d.set("random_y_shift", 0.1)
        d.set("random_x_scaling_range", 0.1)
        d.set("random_y_scaling_range", 0.1)
        d.set("random_x_rotate_range", 10)
        d.set("random_y_rotate_range", 10)
        d.set("random_z_rotate_range", 10)
        d.set("random_distortion", 0.5)
        d.set("random_mixup", 0.1)
        d.set("mixup_alpha", 0.1)
        return d.d

    @staticmethod
    def set_max_cache_size(size):
        pass



#!/usr/bin/env python3

from aimage.head import *
from aimage.img import *
from aimage.ui import *
from aimage.stub_data_loader_od import *
from aimage.stub_data_loader import *

is_native = False
is_available_native_queue = False


def rebuild_worker(n):
    print("stub:rebuild_worker")


def create_queue(n):
    print("stub:create_queue")


def native_decoder(data):
    print("stub:native_decoder")


def native_encoder(data, *, quality=90, format="jpg"):
    print("stub:native_encoder")


def native_fast_decoder(data):
    print("stub:native_fast_decoder")


def native_fast_encoder(data, *, quality=90, format="jpg"):
    print("stub:native_fast_encoder")


def native_load_image(path):
    print("stub:native_load_image")


def native_save_image(path, data, *, quality=90, format="jpg"):
    print("stub:native_save_image")


def native_fast_load_image(path):
    print("stub:native_fast_load_image")


def native_fast_save_image(path, data, *, quality=90, format="jpg"):
    print("stub:native_fast_save_image")


try:
    from aimage_native import *
    print('\033[0;36m' + "========================================================" + '\033[0m')
    print('\033[0;36m' + "Aggressive 3D image augmentation is available." + '\033[0m')
    print('\033[0;36m' + "Fastest async image loader is available." + '\033[0m')
    print('\033[0;36m' + "aimage loading speed is faster than Pillow/OpenCV/TensorFlow." + '\033[0m')
    print('\033[0;36m' + "Event driven non blocking loader is available." + '\033[0m')
    print('\033[0;36m' + "========================================================" + '\033[0m')
    is_native = True
except:
    print('\033[0;31m' + "aimage_native library failed to load." + '\033[0m')
    print('\033[0;31m' + "===========================================================================" + '\033[0m')
    print('\033[0;31m' + "WARN: Native async image library loading failed." + '\033[0m')
    print('\033[0;31m' + " - Pillow/OpenCV/TensorFlow data loading speed is slower than aimage library." + '\033[0m')
    print('\033[0;31m' + " - aimage is superior to the data augmentation systems built into DeepLearning frameworks such as TensorFlow and PyTorch." + '\033[0m')
    print('\033[0;31m' + " If you want to get suport for paid license, please contact support@pegara.com." + '\033[0m')
    print('\033[0;31m' + "===========================================================================" + '\033[0m')
    print()
    print('\033[0;31m' + "Using unoptimized aimage library." + '\033[0m')
    print()
    pass

if __name__ == "__main__":

    pass



#!/usr/bin/env python3
import os
import sys

import acapture
import pyglview

import evaluator

basepath = os.getcwd()
sys.path.append(basepath)

if len(sys.argv) > 1:
    f = sys.argv[1]
cap = acapture.open(f)
view = pyglview.Viewer(keyboard_listener=cap.keyboard_listener)
model = evaluator.Evaluator()


def loop():
    try:
        check, frame = cap.read()
        if check:
            frame = model.render(frame)
            view.set_image(np.array(frame))
    except Exception as e:
        print(e)
        exit(9)

view.set_loop(loop)
view.start()
print("Main thread ended")



#!/usr/bin/env python3
import datetime
import logging
import multiprocessing
import os
import platform
import queue
import re
import shutil
import subprocess
import sys
import threading
import time
import uuid

import numpy as np
import psutil
import twisted.internet.endpoints
import twisted.internet.protocol
import twisted.internet.reactor
import twisted.internet.ssl

if platform.platform() == "Darwin":
    os.environ["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"

logger = logging.getLogger(__name__)
if __name__ == "__main__":
    import protocol as bridge_protocol
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter("%(asctime)s [%(filename)s:%(lineno)d] %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = True
else:
    from . import protocol as bridge_protocol
    logger.addHandler(logging.NullHandler())
    logger.setLevel(logging.DEBUG)
    logger.propagate = True

#################################################################################
# Util

cpu_usage = 0
gpu_usages = []
resource_watcher_queue = queue.Queue()
resource_watcher_thread = None


def resource_watcher(q):
    tm = time.time()
    ROCM_SMI = shutil.which('rocm-smi')
    NVIDIA_SMI = shutil.which('nvidia-smi')
    while True:
        try:
            if time.time() - tm > 2.0:
                if NVIDIA_SMI:
                    try:
                        a = subprocess.check_output("nvidia-smi | grep MiB | grep Default", shell=True).decode("utf-8")
                        if "MiB" in a:
                            index = 0
                            sp = a.split("\n")
                            # canvas = np.zeros((128, 256, 3), dtype=np.uint8)
                            cpu_usage = psutil.cpu_percent()
                            # aimage.draw_text(canvas, "CPU: {}%".format(rpad(str(cpu_usage), 4)), 5, 20 + 20 * index)
                            gpu_usages = []
                            for b in sp:
                                if "Default" in b:
                                    usage = b.split("|")[3].replace("Default", "").strip()
                                    memory = b.split("|")[2].split("/")[0].strip()
                                    # mes = "{} {} {}".format(index, rpad(usage, 4), rpad(memory, 10))
                                    gpu_usages += [[index, usage, memory]]
                                    # aimage.draw_text(canvas, mes, 5, 20 + 20 * (index + 1))
                                    index += 1
                            # aimage.save_image(os.path.join(os.path.dirname(__file__), ".gpu.jpg"), canvas)
                            q.put((cpu_usage, gpu_usages))
                    except Exception as e:
                        pass
                if ROCM_SMI:
                    try:
                        a = subprocess.check_output("rocm-smi | grep Mhz", shell=True).decode("utf-8")
                        if "Mhz" in a:
                            # GPU  Temp   AvgPwr  SCLK    MCLK    Fan     Perf  PwrCap  VRAM%  GPU%
                            # 0    33.0c  18.0W   808Mhz  350Mhz  21.96%  auto  250.0W    0%   0
                            index = 0
                            sp = a.split("\n")
                            # canvas = np.zeros((128, 256, 3), dtype=np.uint8)
                            cpu_usage = psutil.cpu_percent()
                            # aimage.draw_text(canvas, "CPU: {}%".format(rpad(str(cpu_usage), 4)), 5, 20 + 20 * index)
                            gpu_usages = []
                            for b in sp:
                                if "Mhz" in b:
                                    ss = re.findall(r"\S+", b)
                                    gpu_indx = ss[0]
                                    temp = ss[1]
                                    avgpwr = ss[2]
                                    sclk = ss[3]
                                    mclk = ss[4]
                                    fan = ss[5]
                                    perf = ss[6]
                                    pwrcap = ss[7]
                                    vram = ss[8]
                                    gpu_usage = ss[9]
                                    gpu_mem = vram
                                    if len(ss) == 11:
                                        gpu_mem = ss[10]
                                    # mes = "{} {} {}".format(index, rpad(usage, 4), rpad(memory, 10))
                                    gpu_usages += [[gpu_indx, gpu_usage, gpu_mem]]
                                    # aimage.draw_text(canvas, mes, 5, 20 + 20 * (index + 1))
                                    index += 1
                            # aimage.save_image(os.path.join(os.path.dirname(__file__), ".gpu.jpg"), canvas)
                            q.put((cpu_usage, gpu_usages))
                    except Exception as e:
                        pass
                tm = time.time()
            else:
                time.sleep(0.5)
        except Exception as e:
            time.sleep(0.5)


def gpu_log():
    global cpu_usage, gpu_usages, resource_watcher_thread, resource_watcher_queue
    if resource_watcher_thread is None:
        resource_watcher_thread = threading.Thread(name="gpu_log", target=resource_watcher, args=(resource_watcher_queue, ), daemon=True)
        resource_watcher_thread.start()
    try:
        cpu_usage, gpu_usages = resource_watcher_queue.get_nowait()
    except:
        pass
    return cpu_usage, gpu_usages


class StackedServerSocketProtocol(twisted.internet.protocol.Protocol):
    def __init__(self, global_factory, addr):
        super().__init__()
        self.addr = addr
        self.global_factory = global_factory
        self.input_middlewares = []
        self.output_middlewares = []
        self.is_available = False

        self.tm = time.time()
        self.in_ave_q = []
        self.out_ave_q = []
        self.bandwidth_inbound = 0
        self.bandwidth_outbound = 0
        self.total_inbound = 0
        self.total_outbound = 0
        self.uuid = str(uuid.uuid4())
        self.description = ""

    def add_input_protocol(self, p):
        p.queue_name = self.uuid
        self.input_middlewares.append(p)

    def add_output_protocol(self, p):
        p.queue_name = self.uuid
        self.output_middlewares.append(p)

    def connectionMade(self):
        import aimage
        if aimage.is_native:
            cores = multiprocessing.cpu_count()
            aimage.rebuild_worker(cores)
            aimage.create_queue(self.uuid)
        self.is_available = True
        self.global_factory.clients[self.uuid] = self
        #print("C:" + str(self.addr))

    def connectionLost(self, reason):
        import aimage
        if aimage.is_native:
            aimage.delete_queue(self.uuid)
        self.is_available = False
        del self.global_factory.clients[self.uuid]
        #print("D:" + str(self.addr) + str(reason))

    def dataReceived(self, data):
        #logger.debug("TCP:READ:", data[0:10])
        self.bandwidth_inbound += len(data)
        if self.is_available:
            self.input_middlewares[0].write(data)

    def update(self):
        has_event = 0
        if time.time() - self.tm > 1.0:
            if len(self.in_ave_q) > 3:
                self.in_ave_q.pop(0)
                self.out_ave_q.pop(0)
            self.total_inbound += self.bandwidth_inbound
            self.total_outbound += self.bandwidth_outbound
            self.in_ave_q.append(self.bandwidth_inbound)
            self.out_ave_q.append(self.bandwidth_outbound)
            t = datetime.datetime.now()
            ts = f'{t.year}/{t.month}/{t.day} {str(t.hour).zfill(2)}:{str(t.minute).zfill(2)}:{str(t.second).zfill(2)}'
            self.description = "%s://%s:%s %s : I:%.2fMB/s, O:%.2fMB/s TI:%.2fMB, TO:%.2fMB" % (self.addr.type, self.addr.host, self.addr.port, ts, (np.mean(self.in_ave_q) / 1024 / 1024), (np.mean(self.out_ave_q) / 1024 / 1024), self.total_inbound / 1024 / 1024, self.total_outbound / 1024 / 1024)
            self.tm = time.time()
            self.bandwidth_inbound = 0
            self.bandwidth_outbound = 0
        if self.is_available is False: return

        try:
            # From clients
            for i in range(len(self.input_middlewares) - 1):
                b = self.input_middlewares[i].read(-1)
                if len(b):
                    self.input_middlewares[i + 1].write(b)
                    has_event += 1

            for m in self.input_middlewares:
                m.update()

            for i in range(len(self.output_middlewares) - 1):
                b = self.output_middlewares[i].read(-1)
                if len(b):
                    self.output_middlewares[i + 1].write(b)
                    has_event += 1
            for m in self.output_middlewares:
                m.update()

            buf = self.output_middlewares[-1].read(-1)
            if len(buf) > 0:
                self.bandwidth_outbound += len(buf)
                self.transport.write(buf)
                #logger.debug("TCP:WRITE:", buf[0:10])
                has_event += 1
        except Exception as e:
            logger.warning(e)
            try:
                self.transport.close()
                has_event += 1
            except Exception as e:
                logger.warning(e)
                pass
        return has_event

    def read(self, size=-1):
        if self.is_available:
            return self.input_middlewares[-1].read(size)
        return []

    def write(self, objects):
        if self.is_available:
            self.output_middlewares[0].write(objects)


class ObjectTable:
    def __init__(self):
        self.data_table = {}
        pass

    def setDataBlocks(self):
        pass

    def getDataBlocks(self):
        pass


class StreamFactory(twisted.internet.protocol.Factory):
    def __init__(self, **kargs):
        super().__init__()
        self.previous_max_socket_num = 0
        self.enabled_info = False
        self.clients = {}
        self.log_tm = time.time() + 1.0
        self.update()

    def build_protocol_stack(self, s):
        s.add_input_protocol(bridge_protocol.DirectStream())
        s.add_output_protocol(bridge_protocol.DirectStream())

    def buildProtocol(self, addr):
        s = StackedServerSocketProtocol(self, addr)
        s.queue_name = str(uuid.uuid4())
        self.build_protocol_stack(s)
        return s

    def getDataBlocksAsArray(self, size=-1):
        socket_datamap_array = []
        for k in self.clients:
            client_socket = self.clients[k]
            block = client_socket.read()
            if len(block) > 0:
                for data in block:
                    socket_datamap_array.append({"socket": k, "data": data})
        return socket_datamap_array

    def setDataBlocksFromArray(self, socket_datamap_array):
        for obj in socket_datamap_array:
            k = obj["socket"]
            data = obj["data"]
            if k in self.clients:
                client_socket = self.clients[k]
                client_socket.write([data])

    def enable_info(self):
        self.enabled_info = True

    def update(self):
        has_event = 0
        for k in self.clients:
            client_socket = self.clients[k]
            has_event += client_socket.update()
        if self.enabled_info:
            t = time.time()
            if t - self.log_tm > 1.0:
                self.log_tm = t
                cpu_usage, gpu_usages = gpu_log()
                mem = psutil.virtual_memory()
                cpu_res = 'C:{}% '.format(cpu_usage)
                gpu_ress = []
                for g in gpu_usages:
                    gpu_ress += ['G({}:{}:{})'.format(g[0], g[1], g[2])]
                mresources = '[{}M:{:.2f}MB({}%)] [{}]'.format(cpu_res, mem.used / 1024 / 1024, mem.percent, ",".join(gpu_ress))
                print("\033[0K", end="", flush=True)
                print(mresources, flush=True)
                for k in self.clients:
                    print("\033[0K", end="", flush=True)
                    print(self.clients[k].description, flush=True)
                clen = len(self.clients.keys())
                if self.previous_max_socket_num > clen:
                    for i in range(self.previous_max_socket_num - clen):
                        print("\033[0K", end="\n", flush=True)
                for k in range(max(self.previous_max_socket_num, clen)):
                    print("\033[1A", end="\r", flush=True)
                self.previous_max_socket_num = clen
                print("\033[1A", end="\r", flush=True)
        twisted.internet.reactor.callLater(0.001 if has_event > 0 else 0.02, self.update)


# batch_data,src_block,rest_block
def slice_as_batch_size(data_queue, batch_size):
    socket_mapper = []
    batch_data = []
    while len(socket_mapper) < batch_size and len(data_queue) > 0:
        obj = data_queue.pop(0)
        socket_mapper.append(obj)
        batch_data.append(obj["data"])
    return batch_data, socket_mapper, data_queue


def pack_array_datablock(socket_mapper, modified_data):
    dst_mapper = []
    for i in range(len(socket_mapper)):
        obj = socket_mapper[i]
        obj["data"] = modified_data[i]
        dst_mapper.append(obj)
    return dst_mapper


class EaterBridgeServer():
    def getDataBlocksAsArray(self, size=-1):
        try:
            if self.input_queue.empty() is False:
                obj = self.input_queue.get_nowait()
                return obj
        except Exception as e:
            logger.warning(e)
            pass
        return []

    def setDataBlocksFromArray(self, a):
        try:
            if self.output_queue.full() is False:
                self.output_queue.put_nowait(a)
        except Exception as e:
            logger.warning(e)
            return False
        return True

    # def getDataBlocksAsArray(self,size=-1):
    #     obj = self.input_queue.get_nowait()
    #     return self.factory.getDataBlocksAsArray(size)
    # def setDataBlocksFromArray(self,a):
    #     self.factory.setDataBlocksFromArray(a)
    def __init__(self, _dict={}, **kargs):
        for k in kargs:
            _dict[k] = kargs[k]
        kargs = _dict

        if "VSCODE_DEUBG" in os.environ:
            self.input_queue = queue.Queue()
            self.output_queue = queue.Queue()
            self.signal_queue_r = queue.Queue()
            self.signal_queue_w = queue.Queue()
        else:
            self.input_queue = multiprocessing.Queue()
            self.output_queue = multiprocessing.Queue()
            self.signal_queue_r = multiprocessing.Queue()
            self.signal_queue_w = multiprocessing.Queue()

        if "port" not in kargs:
            raise Exception("Required parameter: port was None")
        if "host" not in kargs:
            raise Exception("Required parameter: host was None")

        logger.info("Start server tcp:" + str(kargs["port"]))
        parameter_block = []
        protocol = "tcp"
        if "ssl" in kargs and kargs["ssl"]:
            protocol = "ssl"
            if "key" not in kargs:
                raise Exception("Required parameter: key was None")
            if "crt" not in kargs:
                raise Exception("Required parameter: crt was None")
            parameter_block.append("privateKey=" + kargs["key"])
            parameter_block.append("certKey=" + kargs["crt"])
        parameter_block.append(protocol)
        parameter_block.append(str(kargs["port"]))
        if len(kargs["host"]):
            parameter_block.append("interface=" + kargs["host"].replace(":", "\\:"))

        if "protocol_stack" in kargs:
            self.factory = kargs["protocol_stack"]()
        else:
            raise Exception("protocol_stack is required.")
        parameter = ":".join(parameter_block)
        logger.info(parameter)
        twisted.internet.endpoints.serverFromString(twisted.internet.reactor, parameter).listen(self.factory)

        def runner(input_queue, output_queue, signal_queue_r, signal_queue_w):
            import aimage

            def __update__():
                has_event = 0
                try:
                    if input_queue.full() is False:
                        obj = self.factory.getDataBlocksAsArray(-1)
                        if len(obj) > 0:
                            input_queue.put_nowait(obj)
                            has_event += 1
                except Exception as e:
                    logger.warning(e)
                    pass
                try:
                    if output_queue.empty() is False:
                        obj = output_queue.get_nowait()
                        self.factory.setDataBlocksFromArray(obj)
                        has_event += 1
                    if signal_queue_r.empty() is False:
                        signal_queue_w.put(1)
                        # End of application
                        return
                except Exception as e:
                    logger.warning(e)
                    pass
                twisted.internet.reactor.callLater(0.001 if has_event > 0 else 0.02, __update__)

            __update__()
            twisted.internet.reactor.run(False)

        #runner(self.input_queue, self.output_queue, self.signal_queue_r, self.signal_queue_w)
        #multiprocessing.set_start_method('spawn', True)

        if "VSCODE_DEUBG" in os.environ:
            self.thread = threading.Thread(target=runner, args=(self.input_queue, self.output_queue, self.signal_queue_r, self.signal_queue_w), daemon=True)
        else:
            self.thread = multiprocessing.Process(target=runner, args=(self.input_queue, self.output_queue, self.signal_queue_r, self.signal_queue_w), daemon=True)
        self.thread.start()

    def update(self):
        return 0

    def destroy(self):
        try:
            self.signal_queue_r.put_nowait(None)
            self.signal_queue_w.get()
        except Exception as e:
            logger.warning(e)

    def run(self):
        while True:
            has_event = self.update()
            if has_event == 0:
                time.sleep(0.01)


if __name__ == "__main__":
    from easydict import EasyDict as edict
    import signal

    class ProtocolStack(StreamFactory):
        def build_protocol_stack(self, s):
            s.add_input_protocol(bridge_protocol.LengthSplitIn())
            s.add_input_protocol(bridge_protocol.ImageDecoder())
            s.add_output_protocol(bridge_protocol.ImageEncoder(quality=args.quality))
            s.add_output_protocol(bridge_protocol.LengthSplitOut())
            self.enable_info()

    class Server(EaterBridgeServer):
        def __init__(self, **kargs):
            super().__init__(kargs)
            self.data_queue = []
            self.model = None

        def update(self):
            new_data = self.getDataBlocksAsArray()
            self.data_queue += new_data
            if len(self.data_queue) > 0:
                batch_data, socket_mapper, self.data_queue = slice_as_batch_size(self.data_queue, 4)
                #self.model.eval(batch_data)
                stored_datablocks = pack_array_datablock(socket_mapper, batch_data)
                self.setDataBlocksFromArray(stored_datablocks)
            return len(new_data)

    args = edict()
    args.ssl = False
    args.crt = None
    args.key = None
    args.host = "::0"
    args.port = 4649
    args.quality = 60

    args.protocol_stack = ProtocolStack
    sv = Server(**args.__dict__)

    def terminate(a, b):
        sv.destroy()
        exit(9)

    signal.signal(signal.SIGINT, terminate)
    signal.signal(signal.SIGTERM, terminate)
    sv.run()



#!/usr/bin/env python3
_dopen = open
from aimage.ui import *
from aimage.img import *
from aimage.head import *
import random
import time


class AggressiveImageGeneratorForOD:
    def __init__(self, datas, classes, data_aug_params, shuffle=True):
        self.classes = classes
        self.total = len(datas)
        self.q = 0
        self.iindex = 0
        self.oindex = 0
        self.datas = datas
        self.STREAM_BATCH = 16
        self.data_aug_params = data_aug_params
        self.output_buffer = []
        self.shuffle = shuffle
        self.stub_buffer = []

    def get_data_block(self, batch_size):
        if self.total == 0:
            raise "Zero length"
        while True:
            if self.q < 1024:
                if self.iindex < self.total:
                    ds = self.datas[self.iindex:self.iindex + self.STREAM_BATCH]
                    dlen = len(ds)
                    self.q += dlen
                    stream = list()
                    #input   = dict()
                    #input["data_aug_params"] = self.data_aug_params
                    for d in ds:
                        dd = dict()
                        dd["image_path"] = d["file_path"]
                        dd["image"] = load(d["file_path"])
                        dd["points_table"] = d["bounding_box_table"]
                        stream.append(dd)
                    #input["stream"] = stream
                    # //    {
                    # //      params:{
                    # //          data_aug_params:{}
                    # //      },
                    # //      stream: [{image_path:"",signals:[],points_table:[]},{},{},{},]
                    # //    }
                    # native_module.async_image_loader_with_data_aug_input(str(hex(id(self))),input)
                    self.stub_buffer += stream
                    #self.stub_buffer += [input]
                    # native_module.image_data_augmentation(input)
                    self.iindex += dlen
                else:
                    # random.shuffle(self.datas)
                    # self.iindex = 0
                    break
            else:
                break

        if self.iindex == self.oindex:
            return None
        if len(self.output_buffer) >= batch_size or self.iindex == self.oindex + len(self.output_buffer):
            buf = self.output_buffer[0:batch_size]
            self.output_buffer = self.output_buffer[batch_size:]
            self.oindex += len(buf)
            return buf
        while True:
            #ret = native_module.async_image_loader_with_data_aug_output(str(hex(id(self))))
            ret = self.stub_buffer
            if ret is not None and len(ret) > 0:
                for d in ret:
                    if "points_table" in d:
                        points_table = d["points_table"]
                        del d["points_table"]
                        d["bounding_box_table"] = points_table
                    else:
                        d["bounding_box_table"] = {}
                self.q -= len(ret)
                self.output_buffer += ret
                self.stub_buffer = []
                if len(self.output_buffer) >= batch_size or self.iindex == self.oindex + len(self.output_buffer):
                    buf = self.output_buffer[0:batch_size]
                    self.output_buffer = self.output_buffer[batch_size:]
                    self.oindex += len(buf)
                    return buf
            else:
                time.sleep(0.1)
                if self.iindex == self.oindex:
                    return None

    @staticmethod
    def make_full_aug_params():
        class Dict:
            def __init__(self):
                self.d = dict()

            def set(self, k, v):
                self.d[k] = str(v)

        d = Dict()
        FD = False
        if True:
            d.set("random_crop", 1 if FD else 0.25)  # @@
            d.set("crop_range", 0.2)  # @@
            d.set("random_median", 1 if FD else 0.1)  # @@
            d.set("random_color_reduction", 1 if FD else 0.1)  # @???
            d.set("random_equalization", 1 if FD else 0.1)

            d.set("random_bilateral", 1 if FD else 0.1)
            d.set("random_mosaic", 1 if FD else 0.1)
            d.set("mosaic_shift_range", 1)

            d.set("random_cos", 1 if FD else 0.1)
            d.set("random_hsv", 1 if FD else 0.5)
            d.set("hue_range", 15 if FD else 15)
            d.set("saturation_range", 0.05)
            d.set("lightness_range", 0.1)
            d.set("random_resize_and_arrange_images", 1 if FD else 0.1)
            d.set("random_horizontal_flip", 1 if FD else 0.15)
            d.set("random_vertical_flip", 1 if FD else 0.25)

            d.set("random_gaussian", 1 if FD else 0.1)
            d.set("random_sharp", 1 if FD else 0.1)
            d.set("random_pow", 1 if FD else 0.1)
            d.set("pow_min", 0.9)
            d.set("pow_max", 1.3)
            d.set("random_noise", 1 if FD else 0.1)
            d.set("random_erase", 1 if FD else 0.1)
            d.set("background_type", 3)
            d.set("affine_transform", 1)
            d.set("aggressive_transform", 0)
            d.set("random_z_rotate_range", 3)
            d.set("random_x_shift", 0.1)
            d.set("random_y_shift", 0.1)

        return d.d

    def sync_reset(self):
        if self.shuffle:
            random.shuffle(self.datas)
        self.iindex = 0
        self.oindex = 0

    @staticmethod
    def set_max_cache_size(size):
        pass



#!/usr/bin/env python3

import aimage
import cv2
import os
import numpy as np
import gutil
import json


if __name__ == "__main__":
    img = np.zeros((512,512,3), dtype=np.uint8)
    aimage.draw_fill_rect(img, (0,100), (0, 200))
    aimage.draw_rect(img, (0,100), (0, 200))

    aimage.draw_title(img, "test")
    aimage.draw_footer(img, "test")
    
    print("Done")



#!/usr/bin/env python3
import logging
import struct

import numpy as np

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logger.setLevel(logging.DEBUG)
logger.propagate = True

DEBUG = False



def debug(*args, **kwargs):
    logger.debug(" ".join([str(s) for s in ['\033[1;30m', *args, '\033[0m']]), **kwargs)


def success(*args, **kwargs):
    logger.info(" ".join([str(s) for s in ['\033[0;32m', *args, '\033[0m']]), **kwargs)


def warn(*args, **kwargs):
    logger.warning(" ".join([str(s) for s in ['\033[0;31m', *args, '\033[0m']]), **kwargs)


def info(*args, **kwargs):
    logger.info(" ".join([str(s) for s in ['\033[0;36m', *args, '\033[0m']]), **kwargs)


def check_type(d, name):
    if DEBUG:
        debug(name, type(d))


def to_bytes(d):
    if isinstance(d, bytes):
        return bytearray(d)
    if isinstance(d, bytearray):
        return d
    if isinstance(d, str):
        return d.encode('utf-8')
    if isinstance(d, list):
        return bytearray(d)
    if type(d).__module__ == np.__name__:
        return bytearray(d)
    return False


class StreamIO:
    def __init__(self):
        self.b = bytearray()

    def read(self, size=-1):
        _b = self.b
        if size == -1:
            # Return all
            self.b = bytearray()
            return _b
        # Return sliced buffer
        bb = _b[0:size]
        self.b = _b[size:]
        return bb

    def getbuffer(self):
        return self.b

    def write(self, data):
        _data = to_bytes(data)
        slen = 0
        if _data:
            slen = len(_data)
            self.b.extend(_data)
            return slen
        else:
            ex = "expected type is bytes. Got object was " + str(type(data))
            warn(ex)
            raise ex
        return slen

    def size(self):
        return len(self.b)

    def length(self):
        return len(self.b)


class DirectStream:
    def __init__(self):
        self.queue_name = None
        self.buffer = StreamIO()

    def write(self, data):
        return self.buffer.write(data)

    def read(self, size=-1):
        return self.buffer.read(size)

    def update(self):
        pass

    def info(self):
        return "DirectStream: TCP stream <bytes> => <bytes>"


class LengthSplitIn:  # Stream(socket) to Blocks
    def __init__(self, max_buffer_size=1024 * 1024 * 10):
        self.queue_name = None
        self.buffer = StreamIO()
        self.blocks = []
        self.max_buffer_size = max_buffer_size

    # stream to blocks
    def write(self, data):
        #debug("LengthSplitIn:write:I", data)
        slen = len(data)
        blen = self.buffer.length()
        check_type(data, "W:LengthSplitIn")
        if slen + blen > self.max_buffer_size:
            debug("LengthSplitIn:write", "Data size:", slen, "Buffer size:", blen)
            raise Exception("too much data size")
        self.buffer.write(data)
        buf = self.buffer.getbuffer()
        # More than header size
        while len(buf) >= 4:
            body_length = struct.unpack_from(">I", buf[0:4])[0]
            # Has contents
            if len(buf) >= 4 + body_length:
                head = self.buffer.read(4)
                (head)
                body = self.buffer.read(body_length)
                # debug("LengthSplitIn:write:R", body)
                self.blocks.append(body)
                buf = self.buffer.getbuffer()
            else:
                break
        return slen

    # blocks (extracted)
    def read(self, size=-1):
        if size == -1:
            _blocks = self.blocks
            self.blocks = []
            return _blocks
        _blocks = self.blocks[0:size]
        self.blocks = self.blocks[size:]
        return _blocks

    def update(self):
        pass

    def info(self):
        return "LengthSplitIn: Data block <bytes(<int,bytes,int,bytes,>)> => <[<bytes>,]>"


class LengthSplitOut:  # Block(s) to Stream(socket)
    def __init__(self, max_buffer_size=1024 * 1024 * 10):
        self.queue_name = None
        self.buffer = StreamIO()
        self.max_buffer_size = max_buffer_size

    # single data block to stream
    def write(self, blocks):
        check_type(blocks, "W:LengthSplitOut")

        tlen = 0
        blen = self.buffer.length()
        for data in blocks:
            tlen += len(data)
        if tlen + blen > self.max_buffer_size:
            debug("LengthSplitOut:write", "Data size:", tlen, "Buffer size:", blen)
            raise Exception("too much data size")

        for data in blocks:
            slen = len(data)
            blen = slen.to_bytes(4, 'big')
            self.buffer.write(blen)
            self.buffer.write(data)

        return tlen

    # stream (to socket)
    def read(self, size=-1):
        return self.buffer.read(size)

    def update(self):
        pass

    def info(self):
        return "LengthSplitOut: Data block <[<bytes>,]> => <bytes(<int,bytes,int,bytes,>)>"


class ImageDecoder:  # Blocks to Blocks
    def __init__(self, *, queue_name="default"):
        self.input_blocks = []
        self.processing_map = {}
        self.output_blocks = []
        self.rcv_index = 0
        self.req_index = 0
        self.queue_name = queue_name

    def write(self, blocks):
        check_type(blocks, "W:ImageDecoder")
        slen = 0
        for data in blocks:
            slen += len(data)
            self.input_blocks.append(data)
        return slen

    def read(self, size=-1):
        if size == -1:
            blocks = self.output_blocks
            self.output_blocks = []
            return blocks
        blocks = self.output_blocks[0:size]
        self.output_blocks = self.output_blocks[size:]
        return blocks

    def update(self):
        import aimage
        try:
            if aimage.is_native:
                if aimage.is_available_native_queue:
                    objs = []
                    for b in self.input_blocks:
                        objs.append({"input_buffer": b, "id": self.req_index})
                        self.req_index += 1
                    if len(objs) > 0:
                        aimage.decode_input(objs, self.queue_name)
                    self.input_blocks = []

                    ret = aimage.decode_output(self.queue_name)
                    if len(ret) > 0:
                        for obj in ret:
                            self.processing_map[obj["index"]] = obj
                    while True:
                        obj = self.processing_map.pop(self.rcv_index, None)
                        if obj:
                            self.output_blocks.append(obj["data"])
                            self.rcv_index += 1
                        else:
                            break
                else:
                    # for b in self.input_blocks: self.output_blocks.append(aimage.native_decoder(b))
                    for b in self.input_blocks:
                        self.output_blocks.append(aimage.native_fast_decoder(b))  # x3~x4 faster than OpenCV imdecode loader.
                    self.input_blocks = []
            else:
                # for b in self.input_blocks: self.output_blocks.append(aimage.native_decoder(b))
                for b in self.input_blocks:
                    self.output_blocks.append(aimage.opencv_decoder(b))
                self.input_blocks = []
        except Exception as e:
            warn(e)
            print(e)

    def info(self):
        return "ImageDecoder: Image data block <[<bytes>,]> => <[<ndarray>,]>"


class ImageEncoder:  # Blocks to Blocks
    def __init__(self, *, queue_name="default", quality=90):
        self.input_blocks = []
        self.processing_map = {}
        self.output_blocks = []
        self.req_index = 0
        self.rcv_index = 0
        self.queue_name = queue_name
        self.quality = quality

    def write(self, blocks):
        slen = 0
        check_type(blocks, "W:ImageEncoder")
        if isinstance(blocks, list) is False:
            raise Exception("Arg must be list(numpy,numpy,...)")
        for data in blocks:
            slen += len(data)
            self.input_blocks.append(np.uint8(data))
        return slen

    def read(self, size=-1):
        if size == -1:
            blocks = self.output_blocks
            self.output_blocks = []
            return blocks
        blocks = self.output_blocks[0:size]
        self.output_blocks = self.output_blocks[size:]
        return blocks

    def update(self):
        import aimage
        try:
            if aimage.is_native:
                if aimage.is_available_native_queue:
                    objs = []
                    for b in self.input_blocks:
                        # encode_input:
                        # [
                        #   {
                        #     id: int,
                        #     input_buffer: ndarray,
                        #   },
                        # ]
                        objs.append({"input_buffer": b, "id": self.req_index})
                        # objs.append({"input_buffer": b, "id": self.req_index})
                        self.req_index += 1
                    if len(objs) > 0: aimage.encode_input(objs, self.quality, "jpg", self.queue_name)
                    self.input_blocks = []

                    ret = aimage.encode_output(self.queue_name)

                    # encode_output:
                    # [
                    #   {
                    #     index: int
                    #     data: ndarray,
                    #   },
                    # ]
                    if len(ret) > 0:
                        for obj in ret:
                            self.processing_map[obj["index"]] = obj
                    while True:
                        obj = self.processing_map.pop(self.rcv_index, None)
                        if obj:
                            self.output_blocks.append(obj["data"])
                            self.rcv_index += 1
                        else:
                            break
                else:
                    # for b in self.input_blocks: self.output_blocks.append(aimage.native_encoder(b))
                    for b in self.input_blocks:
                        self.output_blocks.append(aimage.native_fast_encoder(b))  # x3~x4 faster than OpenCV imdecode loader.
                        # self.output_blocks.append(aimage.native_encoder(b, quality=self.quality, format="jpg"))
                    self.input_blocks = []
            else:
                # for b in self.input_blocks: self.output_blocks.append(aimage.native_encoder(b))
                for b in self.input_blocks:
                    self.output_blocks.append(aimage.opencv_encoder(b))
                self.input_blocks = []
        except Exception as e:
            print("Critical:", e)
            warn(e)
            exit(9)

    def info(self):
        return "ImageEncoder: Image data block <[<ndarray>,]> => <[<bytes>,]>"


def protocols():
    debug(DirectStream().info())
    debug(LengthSplitIn().info())
    debug(LengthSplitOut().info())
    debug(ImageDecoder().info())
    debug(ImageEncoder().info())


if __name__ == '__main__':
    debug("=============================================")
    protocols()
    debug("=============================================")



from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))

version = '1.3.9'

install_requires = [
    # 'termios',
    #'opencv-python',
    #'pygame',
    #'PyOpenGL',
    'mss',
    'configparser',
    'imgcat',
    'acapture',
    'twisted',
    'tqdm',
    'numpy',
    'psutil',
    'easydict',
    'service_identity',
]

readme = open("README.md").read()

setup(name='aimage',
    version=version,
    description="Native aimage library wrapper for internal use.",
    long_description="https://github.com/aieater/python_async_image_library\n\n"+readme,
    long_description_content_type='text/markdown',
    classifiers=(
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ),
    keywords='async video image computer vision deeplearning machinelarning augmentation',
    author='Pegara, Inc.',
    author_email='support@pegara.com',
    url='https://github.com/aieater/python_async_image_library',
    license='MIT',
    packages=['aimage','aimage.eater.bridge', 'aimage.eater.application'],
    zip_safe=False,
    install_requires=install_requires,
    entry_points={}
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import signal
import time

import numpy as np

from ..bridge import server as ebs

parser = argparse.ArgumentParser(description='')
parser.add_argument('--inference', type=str, default="demo_object_detection", help='')
parser.add_argument('--ssl', action='store_true', help='')
parser.add_argument('--crt', type=str, default="", help='signed certificate file path')
parser.add_argument('--key', type=str, default="", help='private key file path')
parser.add_argument('--port', type=int, default=3000, help='')
parser.add_argument('--host', type=str, default="localhost", help='')
parser.add_argument('--quality', type=int, default=60, help='')
parser.add_argument('--dataset_dir', type=str, default="man2woman", help='')
args = parser.parse_args()


class ImageServer(ebs.EaterBridgeServer):
    def __init__(self, **kargs):
        super().__init__(**kargs)
        self.data_queue = []
        self.model = None
        print("Load Evaluator")
        import evaluator
        print("Instantiate Evaluator")
        self.model = evaluator.Evaluator()
        print("Evaluator Loaded")

    def update(self):
        new_datas = bridge.getDataBlocksAsArray()
        self.data_queue += new_datas
        if len(self.data_queue) > 0:
            batch_data, socket_mapper, self.data_queue = ebs.slice_as_batch_size(self.data_queue, 128)

            batch_data = np.uint8(batch_data)
            for i in range(len(batch_data)):
                img = batch_data[i]
                r_image = self.model.eval(img)
                batch_data[i] = r_image
            stored_datablocks = ebs.pack_array_datablock(socket_mapper, batch_data)
            bridge.setDataBlocksFromArray(stored_datablocks)

        return len(new_datas)


def server(args):
    return ImageServer(**args.__dict__)



#!/usr/bin/env python3

import math
import mimetypes

import numpy as np
import cv2

CRED = '\033[0;31m'
CCYAN = '\033[0;36m'
CGREEN = '\033[0;32m'
CRESET = '\033[0m'

FONT_TYPE = [cv2.FONT_HERSHEY_COMPLEX_SMALL, cv2.FONT_HERSHEY_SIMPLEX, cv2.FONT_HERSHEY_TRIPLEX]


def gamma(img, g):  # @public
    lookUpTable = np.empty((1, 256), np.uint8)
    for i in range(256):
        lookUpTable[0, i] = np.clip(pow(i / 255.0, g) * 255.0, 0, 255)
    img = cv2.LUT(img, lookUpTable)
    return img


def hue(img, h=0, s=0, v=0):  # @public
    img = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
    if h != 0:
        img[:, :, 1] += h
    if s != 0:
        img[:, :, 1] += s
    if v != 0:
        img[:, :, 2] += v
    img = cv2.cvtColor(img, cv2.COLOR_HSV2RGB)
    return img


def flip(img, t):  # @public
    return np.flip(img, t)


def bchw2bhwc(ts):  # @public
    s = len(ts.size())
    return ts.transpose(s - 3, s - 2).transpose(s - 2, s - 1)


def bhwc2bchw(ts):  # @public
    s = len(ts.size())
    return ts.transpose(s - 2, s - 1).transpose(s - 3, s - 2)


def concat_bhwc_image(ts):
    ts = np.array(ts)
    bsize = len(ts)
    rt = int(math.ceil(math.sqrt(bsize)))
    i = 0
    row = None
    for y in range(rt):
        col = None
        for x in range(rt):
            if i >= bsize:
                break
            b = ts[i]
            if col is not None:
                col = cv2.hconcat([col, b])
            else:
                col = b
            i += 1
        if row is not None:
            try:
                row = cv2.vconcat([row, col])
            except:
                pass
        else:
            row = col
        if i == bsize - 1:
            break

    return row  #np.array(row * 255, dtype=np.uint8)


def rgb2bgr(img):  # @public
    if img.shape[2] != 3:
        raise "src image channel must be 3(RGB)"
    if img.dtype != np.uint8:
        raise "expected dtype is uint8."
    return img[..., ::-1]


def draw_image_alpha(img, img_rgba, sx, sy):  # @public
    print("Does not support alpha channel.")
    return img


def draw_footer(img, message, color=(255, 200, 55), bg=(55, 55, 55), font_scale=1, font_type=0):  # @public
    if isinstance(img, np.ndarray) is False: raise Exception("img must be numpy.")
    if img.dtype != np.uint8: raise Exception("img dtype must be np.uint8.")
    h, w, c = img.shape
    color = [int(c) for c in color]
    font_type = int(font_type)
    cv2.rectangle(img=img, pt1=(0, h), pt2=(w, h - 20), color=bg, thickness=-1)
    cv2.putText(img, message, (5, h - 4), FONT_TYPE[font_type], font_scale, color, 1, lineType=cv2.LINE_AA)


def draw_title(img, message, color=(255, 200, 55), bg=(55, 55, 55), font_scale=1, font_type=0):  # @public
    if isinstance(img, np.ndarray) is False: raise Exception("img must be numpy.")
    if img.dtype != np.uint8: raise Exception("img dtype must be np.uint8.")
    h, w, c = img.shape
    color = [int(c) for c in color]
    font_type = int(font_type)
    cv2.rectangle(img=img, pt1=(0, 0), pt2=(w, 20), color=bg, thickness=-1)
    cv2.putText(img, message, (5, 17), FONT_TYPE[font_type], font_scale, color, 1, lineType=cv2.LINE_AA)


def draw_text(img, message, x=5, y=17, color=(255, 200, 55), font_scale=1, font_type=0):  # @public
    if isinstance(img, np.ndarray) is False: raise Exception("img must be numpy.")
    if img.dtype != np.uint8: raise Exception("img dtype must be np.uint8.")
    color = [int(c) for c in color]
    font_type = int(font_type)
    cv2.putText(img, message, (int(x), int(y)), FONT_TYPE[font_type], font_scale, color, 1, lineType=cv2.LINE_AA)


def is_image_ext(f):  # @public
    e = f.split(".")[-1].lower()
    if e == "jpg":
        return True
    if e == "jpeg":
        return True
    if e == "png":
        return True
    if e == "tiff":
        return True
    if e == "gif":
        return True
    return False


def opencv_decoder(data):
    b = data
    if isinstance(data, np.ndarray):
        nb = data
    elif isinstance(data, bytes):
        nb = np.frombuffer(b, dtype=np.uint8)
    elif isinstance(data, bytearray):
        nb = np.frombuffer(b, dtype=np.uint8)
    else:
        raise Exception("Arg type must be bytes/bytearray/ndarray. Invalid arg:" + type(data))
    data = cv2.imdecode(nb, cv2.IMREAD_COLOR)
    data = data[..., ::-1]

    return data


def opencv_encoder(data, **kargs):
    quality = 90
    if "quality" in kargs:
        quality = kargs["quality"]
    data = data[..., ::-1]
    check, data = cv2.imencode(".jpg", data, [int(cv2.IMWRITE_JPEG_QUALITY), quality])  # quality 1-100
    if check is False:
        raise "Invalid image data"
    return data


def pillow_decoder(data):
    d = np.array(Image.open(BytesIO(data)))
    return d


def pillow_encoder(data, **kargs):
    image = Image.fromarray(data)
    d = BytesIO()
    image.save(d, format="jpeg")
    return d


def _opencv_decoder_(data):
    b = data
    nb = np.asarray(b, dtype=np.uint8)
    data = cv2.imdecode(nb, cv2.IMREAD_COLOR)
    data = data[..., ::-1]
    return data


def _opencv_encoder_(data, **kargs):
    quality = 90
    if "quality" in kargs:
        quality = kargs["quality"]
    data = data[..., ::-1]
    check, data = cv2.imencode(".jpg", data, [int(cv2.IMWRITE_JPEG_QUALITY), quality])  # quality 1-100
    if check is False:
        raise "Invalid image data"
    return data


def native_decoder(data):
    raise Exception("You must install aimage-native-ext")


def native_encoder(data, *, quality=90, format="jpg"):
    raise Exception("You must install aimage-native-ext")


def native_fast_decoder(data):
    raise Exception("You must install aimage-native-ext")


def native_fast_encoder(data, *, quality=90, format="jpg"):
    raise Exception("You must install aimage-native-ext")


def native_load_image(path):
    raise Exception("You must install aimage-native-ext")


def native_save_image(path, data, *, quality=90, format="jpg"):
    raise Exception("You must install aimage-native-ext")


def native_fast_load_image(path):
    raise Exception("You must install aimage-native-ext")


def native_fast_save_image(path, data, *, quality=90, format="jpg"):
    raise Exception("You must install aimage-native-ext")


def decoder(data):  # @public
    return _opencv_decoder_(data)


def encoder(data, **kargs):  # @public
    return _opencv_encoder_(data, **kargs)


def load_image(path):  # @public
    img = cv2.imread(path)
    img = img[..., ::-1]
    return img


def save_image(path, data, *, quality=90, format="jpg"):  # @public
    data = data[..., ::-1]
    return cv2.imwrite(path, data, [cv2.IMWRITE_JPEG_QUALITY, quality])


def load(path):  # @public
    t, ext = mimetypes.guess_type(path)[0].split("/")
    if t == "image":
        img = cv2.imread(path, 3)
        if img is None:
            print(CRED, "\n\nInvalid image file or invalid path. \"%s\"\n\n" % (path, ), CRESET)
            raise "Invalid file or invalid path."
        return img[..., ::-1]
    print(CRED, "\n\nInvalid image file or invalid path. \"%s\"\n\n" % (path, ), CRESET)
    return None


def ratio_resize(img, ww, interpolation="fastest"):  # @public
    s = 1
    if img.shape[0] < img.shape[1]:
        s = ww / img.shape[1]
    else:
        s = ww / img.shape[0]
    w = int(img.shape[1] * s)
    h = int(img.shape[0] * s)
    return cv2.resize(img, (w, h), interpolation=cv2.INTER_AREA)


def crop(img, x, y, x2, y2):  # @public
    return img[x:x2, y:y2]


def resize(img, w, h=None, interpolation="fastest"):  # @public
    if h is None:
        return ratio_resize(img, w, interpolation)
    return cv2.resize(img, (w, h), interpolation=cv2.INTER_AREA)


def draw_rect(img, s, t, c=(255, 0, 0), line=2):  # @public
    cv2.rectangle(img=img, pt1=(int(s[0]), int(s[1])), pt2=(int(t[0]), int(t[1])), color=(int(c[0]), int(c[1]), int(c[2])), thickness=line)


def draw_fill_rect(img, s, t, c=(255, 0, 0)):  # @public
    cv2.rectangle(img=img, pt1=(int(s[0]), int(s[1])), pt2=(int(t[0]), int(t[1])), color=(int(c[0]), int(c[1]), int(c[2])), thickness=-1)


def file_type(d):  # @public
    print("image_head: Does not support API.")
    return None


def image_head(d):  # @public
    img = cv2.imread(d)
    info = {}
    info["height"] = img.shape[0]
    info["width"] = img.shape[1]
    info["channel"] = img.shape[2]
    return info


def generate_colors(C=200):  # @public
    color_table = []
    color_table.append((0, 0, 255))
    color_table.append((0, 255, 0))
    color_table.append((255, 0, 255))
    color_table.append((255, 255, 0))
    color_table.append((0, 255, 255))
    color_table.append((255, 0, 0))
    for c in range(C - len(color_table)):
        CD = 0.1
        TPI = (math.pi * 2) / 3
        TT = 1.123
        d1 = 0.5 + math.cos(CD + TPI + TT * c) * 0.5
        d2 = 0.5 + math.cos(CD + TPI * 2 + TT * c) * 0.5
        d3 = 0.5 + math.cos(CD + TPI * 3 + TT * c) * 0.5

        cc = np.array([d3, d2, d1])

        TT = 1.371
        d1 = 0.5 + math.cos(CD + TPI + TT * c) * 0.5
        d2 = 0.5 + math.cos(CD + TPI * 2 + TT * c) * 0.5
        d3 = 0.5 + math.cos(CD + TPI * 3 + TT * c) * 0.5

        cc = cc + np.array([d1, d2, d3])

        cc = cc * (1.0 - c / C) * 255.0
        cc = np.array(cc, dtype=np.uint8)
        color_table.append((int(cc[0]), int(cc[1]), int(cc[2])))
    return color_table


COLOR_TABLE = generate_colors(1024)  # public


def draw_box(image, box, color, caption=None, fontScale=0.5):  # @public
    if isinstance(image, np.ndarray) is False: raise Exception("image must be numpy.")
    if image.dtype != np.uint8: raise Exception("image dtype must be np.uint8.")
    color = [int(c) for c in color]


    if type(box) == np.ndarray:
        if len(box.shape) == 1 and len(box) == 4:
            pass
        elif len(box.shape) == 2 and box.shape[0] == 2 and box.shape[1] == 2:
            box = box.flatten()
        else:
            raise "Invalid shape."
    elif type(box) == list:
        if len(box) == 2:
            box = np.array([box[0][0], box[0][1], box[1][0], box[1][1]], np.int32)
        else:
            box = np.array([box[0], box[1], box[2], box[3]], np.int32)
    elif type(box) == tuple:
        if len(box) == 2:
            box = np.array([box[0][0], box[0][1], box[1][0], box[1][1]], np.int32)
        else:
            box = np.array([box[0], box[1], box[2], box[3]], np.int32)
    else:
        raise "Invalid type. box => " + type(box)

    box = np.array(box)
    image_h = image.shape[0]
    image_w = image.shape[1]
    box_thick = int(0.6 * (image_h + image_w) / 600.0)
    c1 = (int(box[0]), int(box[1]))
    c2 = (int(box[2]), int(box[3]))
    cr = (int(color[0]), int(color[1]), int(color[2]))
    cv2.rectangle(img=image, pt1=c1, pt2=c2, color=cr, thickness=box_thick)
    if caption:
        t_size = cv2.getTextSize(caption, 0, fontScale, thickness=box_thick // 2)[0]
        c3 = (int(c1[0] + t_size[0]), int(c1[1] - t_size[1] - 3))
        cv2.rectangle(img=image, pt1=c1, pt2=c3, color=cr, thickness=-1)  # filled
        cv2.putText(image, caption, (int(box[0]), int(box[1]) - 2), cv2.FONT_HERSHEY_SIMPLEX, fontScale, (0, 0, 0), box_thick // 2, lineType=cv2.LINE_AA)



#!/usr/bin/env python3

import contextlib
import glob
import importlib
import multiprocessing
import os
import platform
import queue
import sys
import threading
import time

import cv2
import numpy as np


def _ipython_imshow_(image):
    import IPython.display
    image = image[..., ::-1]
    check, img = cv2.imencode(".png", image)
    decoded_bytes = img.tobytes()
    IPython.display.display(IPython.display.Image(data=decoded_bytes))
    return None


def _cshow_(image):
    if is_notebook():
        return _ipython_imshow_(image)
    else:
        import imgcat
        return imgcat.imgcat(image)


def _gshow_(image):
    return _cv2_imshow_("", image)


if importlib.util.find_spec("acapture"):
    import acapture

    class FastJpegCapture(acapture.BaseCapture):
        def __init__(self, fd):
            self.f = os.path.expanduser(fd)
            if self.f[-1] != os.sep:
                self.f += os.sep
            self.f += "**" + os.sep + "*"
            files = glob.glob(self.f, recursive=True)
            self.flist = []
            for f in files:
                filename, ext = os.path.splitext(f)
                ext = ext.lower()
                if ext == ".jpg" or ext == ".jpeg":
                    f = os.path.join(self.f, f)
                    self.flist += [f]

        def is_ended(self):
            return len(self.flist) == 0

        def destroy(self):
            pass

        def read(self):
            while len(self.flist) > 0:
                ff = self.flist.pop(0)
                img = load_image(ff)
                if img is not None:
                    return (True, img)
            return (False, None)

    class AsyncFastJpegCapture(acapture.BaseCapture):
        def __other_process__(self, fd, rq, wq):
            fpath = os.path.expanduser(fd)
            if fpath[-1] != os.sep:
                fpath += os.sep
            fpath += "**" + os.sep + "*"
            files = glob.glob(fpath, recursive=True)
            for f in files:
                filename, ext = os.path.splitext(f)
                ext = ext.lower()
                if ext == ".jpg" or ext == ".jpeg":
                    f = os.path.join(fpath, f)
                    exit_signal = wq.get_nowait()
                    if exit_signal is not None:
                        break
                    img = load_image(f)
                    rq.put_nowait((True, img))
            rq.put_nowait((False, None))

        def __init__(self, fd):
            self.rq = multiprocessing.Queue()
            self.wq = multiprocessing.Queue()
            self.th = multiprocessing.Process(target=self.__other_process__, args=(fd, self.rq, self.wq))
            self.th.start()

        def is_ended(self):
            return len(self.flist) == 0

        def destroy(self):
            pass

        def read(self):
            self.rq.get()

    acapture.DirImgFileStub = AsyncFastJpegCapture
    # acapture.DirImgFileStub = FastJpegCapture

    open = acapture.open  # @public
else:
    print("Could not support video.")
    print("pip3 install pygame acapture")

__front_flag_for_opencv_problem__ = False


def _cv2_imshow_(mes, image):
    global __front_flag_for_opencv_problem__
    image = image[..., ::-1]
    cv2.imshow(mes, image)
    ret = cv2.waitKey(1)
    if __front_flag_for_opencv_problem__ is False:
        __front_flag_for_opencv_problem__ = True
        if platform.system() == "Darwin":
            os.system('''/usr/bin/osascript -e 'tell app "Finder" to set frontmost of process "Python" to true' ''')
            cv2.moveWindow("", 0, 0)

    return ret


def is_notebook():  # @public
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def show(image, console=False):  # @public
    image = np.array(image, dtype=np.uint8)
    if console:
        return _cshow_(image)
    else:
        if "DISPLAY" in os.environ or platform.system() == "Darwin":
            return _gshow_(image)
        else:
            return _cshow_(image)


def wait(w=0):  # @public
    if "DISPLAY" in os.environ or platform.system() == "Darwin":
        return cv2.waitKey(w)
    else:
        time.sleep(w)
        return None


def clear_output():  # @public
    if is_notebook():
        import IPython
        IPython.display.clear_output()
    else:
        print("\033[0;0f")


class _key_observer_:
    try:
        import termios
    except Exception as e:
        print(e)
        print("pip3 install termios")
    def __th_observe_key__(q):
        @contextlib.contextmanager
        def raw_mode(file):
            old_attrs = termios.tcgetattr(file.fileno())
            new_attrs = old_attrs[:]
            new_attrs[3] = new_attrs[3] & ~(termios.ECHO | termios.ICANON)
            try:
                termios.tcsetattr(file.fileno(), termios.TCSADRAIN, new_attrs)
                yield
            finally:
                termios.tcsetattr(file.fileno(), termios.TCSADRAIN, old_attrs)

        with raw_mode(sys.stdin):
            try:
                while True:
                    n = ord(sys.stdin.read(1))
                    q.put(n)
            except Exception as e:
                print("Error:" + str(e))
                pass

    def __init__(self):
        q = queue.Queue()
        thread = threading.Thread(target=_key_observer_.__th_observe_key__, args=(q, ), daemon=True)
        thread.start()
        self.q = q
        self.th = thread

    def get(self):
        try:
            return self.q.get_nowait()
        except:
            return None


def make_key_observer():  # @public
    return _key_observer_()
